/*

    This file is part of the project txt2pho.
    The project is released under the AGPL 3.0 license.
    For details concerning the resulting rights and
    conditions of this license see file License.txt.

*/

using namespace std ;

#if !defined(ASCII_HEADER)
#define ASCII_HEADER


#if 'z' - 'a' + 1 != 26
    #error "Kein ASCII-Zeichensatz"
#endif

#ifdef _Windows
    #include <cstring.h>
    #include <classlib/defs.h>
#else
    #include <string>
#endif


////////////////////////////////////////////////////////////////////////////////////////
//                                                                                    //
//                                   A C H T U N G                                    //
//                                  ---------------                                   //
//                                                                                    //
// Diese Datei enthlt String- und Character-Konstanten mit ASCII- und IPA-Zeichen.   //
// Sie darf unter keinen Umstnden automatisch in einen anderen Zeichensatz           //
// konvertiert werden.                                                                //
//                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////


const int cnt_chars_in_alphabet = 'z' - 'a' + 1;


extern const string asc_lower_chars;
extern const string asc_upper_chars;
extern const string asc_consonants;
extern const string asc_vowels;

// Definitionen fr europische Sonderzeichen

const char asc_a_acc_acute = '';
const char asc_a_acc_grave = '';
const char asc_a_acc_hat   = '';
const char asc_a_umlaut    = '';
const char asc_A_acc_acute = '';
const char asc_A_acc_grave = '';
const char asc_A_acc_hat   = '';
const char asc_A_umlaut    = '';
const char asc_e_acc_acute = '';
const char asc_e_acc_grave = '';
const char asc_e_acc_hat   = '';
const char asc_E_acc_acute = '';
const char asc_E_acc_grave = '';
const char asc_E_acc_hat   = '';
const char asc_i_acc_acute = '';
const char asc_i_acc_grave = '';
const char asc_i_acc_hat   = '';
const char asc_I_acc_acute = '';
const char asc_I_acc_grave = '';
const char asc_I_acc_hat   = '';
const char asc_I_acc_trema = '';
const char asc_o_acc_acute = '';
const char asc_o_acc_grave = '';
const char asc_o_acc_hat   = '';
const char asc_o_umlaut    = '';
const char asc_O_acc_acute = '';
const char asc_O_acc_grave = '';
const char asc_O_acc_hat   = '';
const char asc_O_umlaut    = '';
const char asc_u_acc_acute = '';
const char asc_u_acc_grave = '';
const char asc_u_acc_hat   = '';
const char asc_u_umlaut    = '';
const char asc_U_acc_acute = '';
const char asc_U_acc_grave = '';
const char asc_U_acc_hat   = '';
const char asc_U_umlaut    = '';
const char asc_n_tilde     = '';
const char asc_N_tilde     = '';
const char asc_sz_lig      = '';

////////////////////////////////////////////////////////////////////////////////////////

// Mapping-Routinen

char ucase (char c);
char voll7bit (char c);
char upper_char (char c);
char remap_char (char c);


////////////////////////////////////////////////////////////////////////////////////////

inline bool is_whitespace (char c)
{
    return c == '\0' || c == '\t' || c == '\n' || c == '\r' || c == ' ' || c == 0x0d || c == 0x0a || c == 0x20 ;
}


//

inline bool is_ascii_vowel (char ch)
{
    #ifdef _Windows
    return asc_vowels.contains(ch) ? true : false;
    #else
    return asc_vowels.find(ch) != string::npos;
    #endif
}


inline bool is_ascii_consonant (char ch)
{
    #ifdef _Windows
    return asc_consonants.contains(ch) ? true : false;
    #else
    return asc_consonants.find(ch) != string::npos;
    #endif
}


////////////////////////////////////////////////////////////////////////////////////////

//#if defined(__GNUC__)
char* strrev(char* str);
//char *strdup(const char *str);
//#endif


#endif




